package ru.dbotthepony.mc.otm.core.collect

import it.unimi.dsi.fastutil.HashCommon
import it.unimi.dsi.fastutil.ints.IntBidirectionalIterator
import it.unimi.dsi.fastutil.ints.IntCollection
import it.unimi.dsi.fastutil.ints.IntComparator
import it.unimi.dsi.fastutil.ints.IntIterators
import it.unimi.dsi.fastutil.ints.IntSet
import it.unimi.dsi.fastutil.ints.IntSortedSet

class IntRange2Set private constructor(private val first: Int, private val last: Int) : IntSortedSet {
	constructor(range: IntRange) : this(range.first, range.last) {
		require(range.step == 1) { "Provided range has non-standard step of ${range.step}" }
	}

	@Deprecated("Not supported", level = DeprecationLevel.ERROR)
	override fun add(element: Int): Boolean {
		throw UnsupportedOperationException()
	}

	@Deprecated("Not supported", level = DeprecationLevel.ERROR)
	override fun addAll(c: IntCollection?): Boolean {
		throw UnsupportedOperationException()
	}

	@Deprecated("Not supported", level = DeprecationLevel.ERROR)
	override fun addAll(elements: Collection<Int>): Boolean {
		throw UnsupportedOperationException()
	}

	@Deprecated("Not supported", level = DeprecationLevel.ERROR)
	override fun clear() {
		throw UnsupportedOperationException()
	}

	override fun iterator(fromElement: Int): IntBidirectionalIterator {
		if (isEmpty() || fromElement > last)
			return IntIterators.EMPTY_ITERATOR
		else if (fromElement <= first)
			return IntIterators.fromTo(first, last + 1)
		else
			return IntIterators.fromTo(fromElement, last + 1)
	}

	override fun iterator(): IntBidirectionalIterator {
		if (isEmpty())
			return IntIterators.EMPTY_ITERATOR

		return IntIterators.fromTo(first, last + 1)
	}

	@Deprecated("Not supported", level = DeprecationLevel.ERROR)
	override fun remove(k: Int): Boolean {
		throw UnsupportedOperationException()
	}

	@Deprecated("Not supported", level = DeprecationLevel.ERROR)
	override fun removeAll(c: IntCollection?): Boolean {
		throw UnsupportedOperationException()
	}

	@Deprecated("Not supported", level = DeprecationLevel.ERROR)
	override fun removeAll(elements: Collection<Int>): Boolean {
		throw UnsupportedOperationException()
	}

	@Deprecated("Not supported", level = DeprecationLevel.ERROR)
	override fun retainAll(c: IntCollection?): Boolean {
		throw UnsupportedOperationException()
	}

	@Deprecated("Not supported", level = DeprecationLevel.ERROR)
	override fun retainAll(elements: Collection<Int>): Boolean {
		throw UnsupportedOperationException()
	}

	override fun contains(key: Int): Boolean {
		return key in first .. last
	}

	override fun containsAll(c: IntCollection): Boolean {
		return c.ktIterator().all { it in first .. last }
	}

	override fun containsAll(elements: Collection<Int>): Boolean {
		return elements.all { it in first .. last }
	}

	override fun isEmpty(): Boolean {
		return last > first
	}

	override fun toArray(a: IntArray?): IntArray {
		if (a == null || a.size < size)
			return toIntArray()

		var index = 0

		for (i in first .. last) {
			a[index++] = i
		}

		return a
	}

	override fun toIntArray(): IntArray {
		if (isEmpty())
			return EMPTY_ARRAY

		return IntArray(last - first + 1) { first + it }
	}

	override fun comparator(): IntComparator? {
		return null
	}

	override fun subSet(fromElement: Int, toElement: Int): IntSortedSet {
		if (fromElement <= first && toElement > last)
			return this
		else if (fromElement <= first)
			return IntRange2Set(first, toElement - 1)
		else if (toElement > last)
			return IntRange2Set(fromElement, last)
		else
			return EMPTY
	}

	override fun headSet(toElement: Int): IntSortedSet {
		if (isEmpty() || toElement <= first)
			return EMPTY
		else if (toElement < last)
			return this
		else
			return IntRange2Set(first, toElement - 1)
	}

	override fun tailSet(fromElement: Int): IntSortedSet {
		if (isEmpty() || fromElement > last)
			return EMPTY
		else if (fromElement < first)
			return this
		else
			return IntRange2Set(fromElement, last)
	}

	override fun firstInt(): Int {
		if (isEmpty())
			throw NoSuchElementException("Range is empty")

		return first
	}

	override fun lastInt(): Int {
		if (isEmpty())
			throw NoSuchElementException("Range is empty")

		return last
	}

	override val size: Int
		get() = if (last > first) 0 else last - first + 1

	override fun toString(): String {
		if (isEmpty())
			return "IntRange2Set[EMPTY]"

		return "IntRange2Set[$first .. $last]"
	}

	override fun equals(other: Any?): Boolean {
		return this === other ||
			other is IntRange2Set && (isEmpty() == other.isEmpty() || first == other.first && last == other.last) ||
			other is IntSet && other.size == size && containsAll(other) ||
			other is Set<*> && other.size == size && other.all { it is Int && it in first .. last }
	}

	override fun hashCode(): Int {
		if (isEmpty())
			return EMPTY_HASH

		return HashCommon.murmurHash3(last - first)
	}

	companion object {
		/**
		 * Returns set containing values beginning from [from] (inclusive) to [to] (inclusive)
		 */
		fun closed(from: Int, to: Int): IntRange2Set {
			if (from > to)
				return EMPTY

			return IntRange2Set(from, to)
		}

		/**
		 * Returns set containing values beginning from [from] (inclusive) to [to] (exclusive)
		 */
		fun openEnded(from: Int, to: Int): IntRange2Set {
			return closed(from, to - 1)
		}

		private val EMPTY_ARRAY = IntArray(0)
		val EMPTY = IntRange2Set(0, -1)
		private val EMPTY_HASH = HashCommon.murmurHash3(0)
	}
}
